* SUBROUTINE UXDLGF                ALL SYSTEMS                91/12/01
C PORTABILITY : ALL SYSTEMS
C 91/12/01 MA : ORIGINAL VERSION
*
* PURPOSE :
* FACTORIZATION A+E=P*U*D*TRANS(U)*TRANS(P) OF A DENSE SYMMETRIC
* POSITIVE DEFINITE MATRIX A+E WHERE P IS A PERMUTATION MATRIX, D AND E
* ARE DIAGONAL POSITIVE DEFINITE MATRICES AND U IS AN UPPER TRIANGULAR
* MATRIX. IF A IS SUFFICIENTLY POSITIVE DEFINITE THEN E=0.
*
* PARAMETERS :
*  II  N ORDER OF THE MATRIX A.
*  RU  A(N*(N+1)/2)  ON INPUT A GIVEN DENSE SYMMETRIC (USUALLY POSITIVE
*         DEFINITE) MATRIX A STORED IN THE PACKED FORM. ON OUTPUT THE
*         COMPUTED FACTORIZATION A+E=P*U*D*TRANS(U)*TRANS(P).
*  IO  IA(N)  A PERMUTATION FIELD WHICH CONTAINS AN INFORMATION ABOUT
*         PERMUTATION MATRIX P.
*  RA  B(N)  AUXILIARY VECTOR.
*  IO  INF  AN INFORMATION OBTAINED IN THE FACTORIZATION PROCESS. IF
*         INF=0 THEN A IS SUFFICIENTLY POSITIVE DEFINITE AND E=0. IF
*         INF<0 THEN A IS NOT SUFFICIENTLY POSITIVE DEFINITE AND E>0. IF
*         INF>0 THEN A IS INDEFINITE AND INF IS AN INDEX OF THE
*         MOST NEGATIVE DIAGONAL ELEMENT USED IN THE FACTORIZATION
*         PROCESS.
*  RU  SIG  ON INPUT A DESIRED TOLERANCE FOR POSITIVE DEFINITENESS. ON
*         OUTPUT THE MOST NEGATIVE DIAGONAL ELEMENT USED IN THE
*         FACTORIZATION PROCESS (IF INF>0).
*  RO  TAU  MAXIMUM DIAGONAL ELEMENT OF THE MATRIX E.
*  IA  Q(N) AUXILIARY VECTOR.
*
* METHOD :
* R.B.SCHNABEL, E.ESKOW: A NEW CHOLESKI FACTORIZATION, SIAM J.
* SCI. STAT. COMPUT. 11 (1990), PP. 1136-1158.
*
      SUBROUTINE UXDLGF(N,A,IA,B,INF,SIG,TAU,Q)
      INTEGER N,IA(N),INF,Q(N)
      REAL*8  A(N*(N+1)/2),B(N),SIG,TAU
      REAL*8  ALF,BET,DEL,GAM,RHO,TOL,R1,TEMP
      INTEGER  I,ID,J,JD,K,KC,KD,KD1,KP,KPC,KX
      REAL*8 ZERO,ONE,TWO,HALF
      PARAMETER (ZERO=0.0D 0,ONE=1.0D 0,TWO=2.0D 0,HALF=0.5D 0)
C
      INF = 0
      TOL = SIG
      SIG = ZERO
      TAU = ZERO
C
C     ESTIMATION OF THE MATRIX NORM
C
      GAM = TOL
      KD = 0
      DO K = 1, N
          KD = KD + K
          GAM = MAX(GAM,ABS(A(KD)))
          Q(K) = K
      ENDDO
C
C     PHASE ONE, A POTENTIALLY POSITIVE DEFINITE
C
      K   = N           ! THE MAIN COLUMN INDEX
      KD1 = (N-1)*N/2   ! INDEX OF A(K-1,K-1)
      KC  = KD1 + 1     ! INDEX OF A(1,K)
      KD  = KD1 + K     ! INDEX OF A(K,K)
   10 CONTINUE
C
C     PIVOT ON MAXIMUM DIAGONAL OF REMAINING SUBMATRIX
C
      ALF = A(KD)   ! MAXIMUM DIAGONAL ELEMENT IN MATRIX A
      KP = K          ! INDEX OF MAXIMUM DIAGONAL ELEMENT
      JD = KD1
      DO J = K-1, 1, -1
          IF (A(JD).GT.ALF) THEN
              ALF = A(JD)
              IF (ALF.GT.ZERO) KP = J
          ENDIF
          JD = JD - J
      ENDDO
      IF (K.NE.KP) CALL UNSWPI(Q(K),Q(KP))
      IF (ALF.LE.ZERO) GOTO 20
      IF (KP.NE.K) THEN
C
C         SWITCH ROWS AND COLUMNS KP AND K
C
          KPC = (KP-1)*KP / 2 + 1
          CALL UXVSWP(KP-1,A(KC),A(KPC),1)
          KX = KPC + KP - 1
          DO J = KP+1, K-1
              KX = KX + J - 1
              CALL UNSWAP( A(KC+J-1), A(KX) )
          END DO
          CALL UNSWAP( A(KC+K-1), A(KPC+KP-1) )
      ENDIF
C
C     TEST ON POSITIVE DEFINITENESS
C
      JD = KD1
      DO J = K-1, 1, -1
          IF (A(JD)-A(KD1+J)**2/ALF.LT.TOL*GAM) GOTO 20
          JD = JD - J
      ENDDO
C
C     PERFORM K-TH ITERATION OF FACTORIZATION
C
      R1 = ONE / ALF
      CALL UXDSMU(K-1,A,-R1,A(KC))
C
C     STORE U(K) IN COLUMN K
C
      CALL UXVSCL(K-1,R1,A(KC),A(KC))
C
C     DECREASE K AND RETURN TO THE START OF THE MAIN LOOP
C
      KD  = KD1
      K   = K - 1
      KC  = KC - K
      KD1 = KD - K
      IF (K.GE.1) GOTO 10
      GOTO 90
C
C     PHASE TWO, A NOT POSITIVE DEFINITE
C
   20 CONTINUE
      IF (K.LE.2) GOTO 50
C
C     CALCULATE LOWER GERSCHGORIN BOUND OF REMAINING SUBMATRIX
C
      JD = KD
      DO J = K, 1, -1
          B(J) = A(JD)
          DO I = J-1, 1, -1
              B(J) = B(J) - ABS(A(JD-I))
          ENDDO
          ID = JD
          DO I = J+1, K
              B(J) = B(J) - ABS(A(ID+J))
              ID = ID + I
          ENDDO
          JD = JD - J
      ENDDO
C
C     MODIFIED CHOLESKI DECOMPOSITION
C
   30 CONTINUE
C
C     PIVOT ON MAXIMUM LOWER GERSCHGORIN BOUND ESTIMATE
C
      TEMP = B(1)
      KP = 1
      DO J=2,K
        IF (B(J).GE.TEMP) THEN
            TEMP = B(J)
            KP = J
        ENDIF
      ENDDO
      IF (KP.NE.K) THEN
C
C         SWITCH ROWS AND COLUMNS KP AND K
C
          CALL UNSWPI(Q(K),Q(KP))
          KPC = (KP-1)*KP / 2 + 1
          CALL UXVSWP(KP-1,A(KC),A(KPC),1)
          KX = KPC + KP - 1
          DO J = KP+1, K-1
              KX = KX + J - 1
              CALL UNSWAP( A(KC+J-1), A(KX) )
          END DO
          CALL UNSWAP( A(KC+K-1), A(KPC+KP-1) )
      ENDIF
C
C     CALCULATE K-TH ELEMENT OF THE CORRECTION MATRIX
C
      BET = ZERO
      DO I = 1, K-1
          BET = BET + ABS(A(KD-I))
      ENDDO
      ALF = A(KD)
      IF (SIG.GT.ALF) THEN
          SIG = ALF
          INF = K
      ENDIF
      RHO=MAX(ZERO,-ALF+MAX(BET,TOL*GAM),TAU)
      IF (RHO.GT.ZERO) THEN
          ALF = ALF + RHO
          A(KD) = ALF
          TAU = RHO
      ENDIF
      IF (ALF.NE.BET) THEN
C
C         UPDATE GERSCHGORIN BOUND ESTIMATE
C
          BET = ONE - BET / ALF
          JD = KD1
          DO I = 1, K-1
              B(I) = B(I) + ABS(A(JD+I))*BET
          ENDDO
      ENDIF
C
C     PERFORM K-TH ITERATION OF FACTORIZATION
C
      R1 = ONE / ALF
      CALL UXDSMU(K-1,A,-R1,A(KC))
C
C     STORE U(K) IN COLUMN K
C
      CALL UXVSCL(K-1,R1,A(KC),A(KC))
C
C     DECREASE K AND RETURN TO THE START OF THE MAIN LOOP
C
      KD  = KD1
      K   = K - 1
      KC  = KC - K
      KD1 = KD - K
      IF (K.GT.2) GOTO 30
C
   50 CONTINUE
      IF (K.EQ.1) GOTO 60
C
C     FINAL 2 X 2 SUBMATRIX
C
      ALF=A(3)
      IF (SIG.GT.ALF) THEN
          SIG = ALF
          INF = 2
      ENDIF
      DEL = A(1)
      BET = HALF*(DEL+ALF)
      DEL = HALF*(DEL-ALF)
      DEL = SQRT(DEL**2+A(2)**2)
      BET = BET - DEL
      RHO = MAX(ZERO,-BET+TOL*MAX(TWO*DEL/(ONE-TOL),GAM))
      IF (RHO.GT.ZERO) THEN
          A(3) = ALF + RHO
          TAU = MAX(RHO,TAU)
      ENDIF
      DEL = A(2)
      A(2) = DEL / A(3)
      ALF = A(1) - A(2)*DEL
      IF (SIG.GT.ALF) THEN
          SIG = ALF
          INF = 1
      ENDIF
      A(1) = ALF + RHO
      GOTO 90
C
   60 CONTINUE
C
C     FINAL 1 X 1 SUBMATRIX
C
      ALF = A(1)
      IF (SIG.GT.ALF) THEN
          SIG = ALF
          INF = 1
      ENDIF
      RHO = MAX(ZERO,-ALF+TOL*GAM)
      IF (RHO.GT.ZERO) THEN
          A(1) = ALF + RHO
          TAU = MAX(RHO,TAU)
      ENDIF
      IF (INF.EQ.0.AND.TAU.GT.ZERO) INF = -1
C
   90 CONTINUE
C
C     PERMUTATION FIELD IA FROM PERMUTATION MATRIX Q
C
      DO K=N,2,-1
          KP = K
   91     KX = Q(KP)
          IF (KX.LE.K) THEN
              IA(K) = KX
          ELSE
              KP = KX
              GOTO 91
          ENDIF
      ENDDO
      IA(1) = 1
      RETURN
      END
