* SUBROUTINE UXDSGF                ALL SYSTEMS                91/12/01
C PORTABILITY : ALL SYSTEMS
C 91/12/01 LU : ORIGINAL VERSION
*
* PURPOSE :
* FACTORIZATION A+E=P*L*D*TRANS(L)*TRANS(P) OF A DENSE SYMMETRIC
* POSITIVE DEFINITE MATRIX A+E WHERE P IS PERMUTATION MATRIX, D AND E
* ARE DIAGONAL POSITIVE DEFINITE MATRICES AND L IS A LOWER TRIANGULAR
* MATRIX. IF A IS SUFFICIENTLY POSITIVE DEFINITE THEN E=0.
*
* PARAMETERS :
*  II  N ORDER OF THE MATRIX A.
*  RU  A(N*(N+1)/2)  ON INPUT A GIVEN DENSE SYMMETRIC (USUALLY POSITIVE
*         DEFINITE) MATRIX A STORED IN THE PACKED FORM. ON OUTPUT THE
*         COMPUTED FACTORIZATION A+E=P*L*D*TRANS(L)*TRANS(P).
*  IO  IA(N)  A PERMUTATION FIELD WHICH CONTAINS AN INFORMATION ABOUT
*         PERMUTATION MATRIX P.
*  RA  B(N)  AUXILIARY VECTOR.
*  IO  INF  AN INFORMATION OBTAINED IN THE FACTORIZATION PROCESS. IF
*         INF=0 THEN A IS SUFFICIENTLY POSITIVE DEFINITE AND E=0. IF
*         INF<0 THEN A IS NOT SUFFICIENTLY POSITIVE DEFINITE AND E>0. IF
*         INF>0 THEN A IS INDEFINITE AND INF IS AN INDEX OF THE
*         MOST NEGATIVE DIAGONAL ELEMENT USED IN THE FACTORIZATION
*         PROCESS.
*  RU  SIG  ON INPUT A DESIRED TOLERANCE FOR POSITIVE DEFINITENESS. ON
*         OUTPUT THE MOST NEGATIVE DIAGONAL ELEMENT USED IN THE
*         FACTORIZATION PROCESS (IF INF>0).
*  RO  TAU  MAXIMUM DIAGONAL ELEMENT OF THE MATRIX E.
*
* METHOD :
* R.B.SCHNABEL, E.ESKOW: A NEW CHOLESKI FACTORIZATION, SIAM J.
* SCI. STAT. COMPUT. 11 (1990), PP. 1136-1158.
*
      SUBROUTINE UXDSGF(N,A,IA,B,INF,SIG,TAU)
      INTEGER N,IA(N),INF
      REAL*8  A(N*(N+1)/2),B(N),SIG,TAU
      REAL*8  ALF,BET,GAM,DEL,RHO,TOL
      INTEGER  I,J,K,L,IP,JP,KP,II,IJ,JJ,IK,JK,KK,IND
      REAL*8 ZERO,ONE,TWO,HALF
      PARAMETER (ZERO=0.0D 0,ONE=1.0D 0,TWO=2.0D 0,HALF=0.5D 0)
      IND(I,J)=MAX(I,J)*(MAX(I,J)-1)/2+MIN(I,J)
      INF=0
      TOL=SIG
C
C     ESTIMATION OF THE MATRIX NORM
C
      GAM=TOL
      SIG=ZERO
      TAU=ZERO
      KK=0
      DO 1 K=1,N
      IA(K)=K
      KK=KK+K
      GAM=MAX(GAM,ABS(A(KK)))
    1 CONTINUE
C
C     PHASE ONE, A POTENTIALLY POSITIVE DEFINITE
C
      DO 6 K=1,N
C
C     PIVOT ON MAXIMUM DIAGONAL OF REMAINING SUBMATRIX
C
      ALF=ZERO
      DO 2 J=K,N
      JP=IA(J)
      JJ=IND(JP,JP)
      BET=A(JJ)
      IF (ALF.LT.BET) THEN
      I=J
      KP=JP
      KK=JJ
      ALF=BET
      ENDIF
    2 CONTINUE
      IF (ALF.LE.ZERO) GO TO 7
      IF (I.NE.K) THEN
      IA(I)=IA(K)
      IA(K)=KP
      ENDIF
C
C     TEST ON POSITIVE DEFINITENESS
C
      DO 3 J=K+1,N
      JP=IA(J)
      JJ=IND(JP,JP)
      JK=IND(JP,KP)
      BET=A(JJ)-(A(JK)/ALF)*A(JK)
      IF (BET.LT.TOL*GAM) GO TO 7
    3 CONTINUE
C
C     PERFORM K-TH ITERATION OF FACTORIZATION
C
      DO 5 J=K+1,N
      JP=IA(J)
      JJ=IND(JP,JP)
      JK=IND(JP,KP)
      DEL=A(JK)
      A(JK)=DEL/ALF
      DO 4 I=K+1,J
      IP=IA(I)
      IJ=IND(IP,JP)
      IK=IND(IP,KP)
      A(IJ)=A(IJ)-A(IK)*DEL
    4 CONTINUE
    5 CONTINUE
    6 CONTINUE
      RETURN
C
C     PHASE TWO, A NOT POSITIVE DEFINITE
C
    7 L=K
      IF (L.LE.N-2) THEN
C
C     CALCULATE LOWER GERSCHGORIN BOUND OF REMAINDING SUBMATRIX
C
C
      DO 9 J=L,N
      JP=IA(J)
      JJ=IND(JP,JP)
      B(JP)=A(JJ)
      DO 8 I=L,N
      IF (I.NE.J) THEN
      IP=IA(I)
      IJ=IND(IP,JP)
      B(JP)=B(JP)-ABS(A(IJ))
      ENDIF
    8 CONTINUE
    9 CONTINUE
C
C     MODIFIED CHOLESKI DECOMPOSITION
C
      DO 15 K=L,N-2
      KP=IA(K)
C
C     PIVOT ON MAXIMUM LOWER GERSCHGORIN BOUND ESTIMATE
C
      I=K
      IP=KP
      BET=B(KP)
      DO 10 J=K+1,N
      JP=IA(J)
      IF (B(JP).GT.BET) THEN
      I=J
      IP=JP
      BET=B(JP)
      ENDIF
   10 CONTINUE
      IF (I.NE.K) THEN
      IA(I)=KP
      IA(K)=IP
      KP=IP
      ENDIF
      KK=IND(KP,KP)
C
C     CALCULATE K-TH ELEMENT OF THE CORRECTION MATRIX
C
      BET=ZERO
      DO 11 J=K+1,N
      JP=IA(J)
      JK=IND(JP,KP)
      BET=BET+ABS(A(JK))
   11 CONTINUE
      ALF=A(KK)
      IF (SIG.GT.ALF) THEN
      SIG=ALF
      INF=K
      ENDIF
      RHO=MAX(ZERO,-ALF+MAX(BET,TOL*GAM),TAU)
      IF (RHO.GT.ZERO) THEN
      ALF=ALF+RHO
      A(KK)=ALF
      TAU=RHO
      ENDIF
      IF (ALF.NE.BET) THEN
C
C     UPDATE GERSCHGORIN BOUND ESTIMATE
C
      BET=ONE-BET/ALF
      DO 12 J=K+1,N
      JP=IA(J)
      JK=IND(JP,KP)
      B(JP)=B(JP)+ABS(A(JK))*BET
   12 CONTINUE
      ENDIF
C
C     PERFORM K-TH ITERATION OF FACTORIZATION
C
      DO 14 J=K+1,N
      JP=IA(J)
      JJ=IND(JP,JP)
      JK=IND(JP,KP)
      DEL=A(JK)
      A(JK)=DEL/ALF
      DO 13 I=K+1,J
      IP=IA(I)
      IJ=IND(IP,JP)
      IK=IND(IP,KP)
      A(IJ)=A(IJ)-A(IK)*DEL
   13 CONTINUE
   14 CONTINUE
   15 CONTINUE
      ENDIF
      IF (L.LE.N-1) THEN
C
C     FINAL 2 X 2 SUBMATRIX
C
      IP=IA(N-1)
      JP=IA(N)
      II=IND(IP,IP)
      IJ=IND(IP,JP)
      JJ=IND(JP,JP)
      ALF=A(II)
      IF (SIG.GT.ALF) THEN
      SIG=ALF
      INF=N-1
      ENDIF
      DEL=A(JJ)
      BET=HALF*(DEL+ALF)
      DEL=HALF*(DEL-ALF)
      DEL=SQRT(DEL**2+A(IJ)**2)
      BET=BET-DEL
      RHO=MAX(ZERO,-BET+TOL*MAX(TWO*DEL/(ONE-TOL),GAM))
      IF (RHO.GT.ZERO) THEN
      A(II)=ALF+RHO
      TAU=MAX(RHO,TAU)
      ENDIF
      DEL=A(IJ)
      A(IJ)=DEL/A(II)
      ALF=A(JJ)-A(IJ)*DEL
      IF (SIG.GT.ALF) THEN
      SIG=ALF
      INF=N
      ENDIF
      A(JJ)=ALF+RHO
      ELSE
C
C     FINAL 1 X 1 SUBMATRIX
C
      IP=IA(N)
      II=IND(IP,IP)
      ALF=A(II)
      IF (SIG.GT.ALF) THEN
      SIG=ALF
      INF=N
      ENDIF
      RHO=MAX(ZERO,-ALF+TOL*GAM)
      IF (RHO.GT.ZERO) THEN
      A(II)=ALF+RHO
      TAU=MAX(RHO,TAU)
      ENDIF
      ENDIF
      IF (INF.EQ.0.AND.TAU.GT.ZERO) INF=-1
      RETURN
      END
